;****************************************************************************
; BIOS patch code for
; - ASUS P2B-F (with hardware monitor) 1014 beta 003 (1014f.003)
; - ASUS P2B-F (without hardware monitor) 1014 beta 003 (1014fn.003)
; - ASUS P2B (with hardware monitor) 1014 beta 003 (1014.003)
; - ASUS P2B (without hardware monitor) 1014 beta 003 (1014nh.003)
; - ASUS P2B-D 1014 beta 003 (1014d.003)
; - ASUS P2B-DS 1014 beta 003 (1014ds.003)
; - ASUS P2L97 (with hardware monitor) 1010 beta 004 (1010.004)
; - ASUS P2L97 (without hardware monitor) 1010 beta 004 (1010m.004)
; - ASUS P2L97-S 1010 beta 004 (1010S.004)
; - ASUS P2-99 (with hardware monitor) 1013 beta 005 (1013.005)
; - ASUS P2-99 (without hardware monitor) 1013 beta 005 (1013d9.005)
; - ASUS P2B-N (with ATI Rage IIC) 1014 beta 003 (1014niic.003)
; - ASUS P2B-N (with ATI Rage Pro) 1014 beta 003 (1014npro.003)
; - ASUS P2L97A 1010 beta 004 (1010a.004)
;****************************************************************************
;
; Main file containing new code and data to be added to the BIOS
;
;****************************************************************************
;
; Changes:
; - support new multipilers for CPU speed/FSB calculation
; - fix for the display of processor brand names at bootup, including support
;   for Pentium III-S and Pentium Pro processors
; - fix for the display of frequency suffixes (A, E, B, EB) at bootup
;   according to Intel datasheets
; - support of additional voltage ranges for processors with Tualatin core
;   and Pentium Pro processors to suppress Hardware Monitor errors at bootup
; - added Setup Utility "Enable" option for Symbios SCSI BIOS to force
;   loading the ROM module even if no matching adapter is present
; - experimental support for Via C3 processors
; - disabled -5 V display for modern PSUs without -5 V rail
; - changed fan speed range to accommodate slow-spinning fans
;
;****************************************************************************

;----------------------------------------------------------------------------
; assembler directives/options
;----------------------------------------------------------------------------

[map symbols bin\mainE000.map]
bits 16
cpu 686

;----------------------------------------------------------------------------
; includes
;----------------------------------------------------------------------------

%include	"macro.def"		; macro definitions
%include	"bios.equ"		; addresses of BIOS code and data
					; used in patch routines

;----------------------------------------------------------------------------
; origin
;----------------------------------------------------------------------------

patchStart:
	org 	patchLocE000		; go to free location in E000 segment
					; to place patch data and code into

;----------------------------------------------------------------------------
; data
;----------------------------------------------------------------------------

ManufacturerIDs:
	dw	IntelTypeMap		; Intel
	db	'Genu'
	db	'ntel'
	db	'ineI'
	dw	ViaTypeMap		; VIA
	db	'Cent'
	db	'auls'
	db	'aurH'
	dw	0FFFFh			; terminator

IntelTypeMap:				; mapping of Intel processor CPUIDs
					; to BIOS-internal "types"
	dw	0630h, 02h		; Klamath core
	dw	0650h, 03h		; Deschutes core
					; (Covington core with same CPUID is
					; labeled as "type 1" via code)
	dw	0660h, 04h		; Mendocino core
	dw	0670h, 05h		; Katmai core
	dw	0680h, 06h		; Coppermine core
	dw	0690h, 07h		; Banias core
					; (not relevant, since only mobile
					; processors with this core exist)
	dw	06B0h, 08h		; Tualatin core
	dw	0610h, 09h		; Pentium Pro
	dw	1630h, 0Ah		; Pentium II Overdrive (uses a
					; Deschutes 065x core)
	dw	0FFFFh			; terminator

ViaTypeMap:
	dw	0660h, 40h		; mapping of Via processor CPUIDs to
					; BIOS-internal "types"
	dw	0670h, 41h		; cores from 0678h onwards are mapped
					; to "type 42h" via code
	dw	0680h, 43h
	dw	0690h, 44h
	dw	0FFFFh			; terminator

%ifndef SMP
aPentiumPro:
	db	'Intel(R) Pentium(R) Pro', 0
aPentium2:
	db	'Intel(R) Pentium(R) II', 0
aPentium3:
	db	'Intel(R) Pentium(R) III', 0
aPentium3S:
	db	'Intel(R) Pentium(R) III-S', 0
%else					; SMP
aPentiumPro:
	db	'Pentium Pro', 0
aCeleron:
	db	'Celeron', 0
aPentium3S:
	db	'Pentium III-S', 0
%endif					; SMP
aViaC3:
	db	'VIA C3', 0

SpeedTable:
	dw	2133
	db	11010000b
	dw	2066
	db	11010000b
	dw	2000
	db	11010000b
	dw	1933
	db	11010000b
	dw	1866
	db	11010000b
	dw	1800
	db	11010000b
	dw	1733
	db	11010000b
	dw	1666
	db	11010000b
	dw	1600
	db	10010000b
	dw	1550
	db	10010000b
	dw	1533
	db	11010000b
	dw	1500
	db	10010000b
	dw	1466
	db	11010000b
	dw	1450
	db	10010000b
	dw	1400
	db	10010000b
	dw	1350
	db	10010000b
	dw	1333
	db	11010000b
	dw	1300
	db	10010000b
	dw	1266
	db	11010000b
	dw	1250
	db	10010000b
	dw	1200
	db	10010000b
	dw	1150
	db	10010000b
	dw	1133
	db	11010000b
	dw	1100
	db	10010000b
	dw	1066
	db	11010000b
	dw	1000
	db	10010000b
	dw	950
	db	10010000b
	dw	933
	db	11010000b
	dw	900
	db	10010000b
	dw	866
	db	11010000b
	dw	850
	db	10010000b
	dw	833
	db	01000000b
	dw	800
	db	10010000b
	dw	766
	db	01000000b
	dw	750
	db	10010000b
	dw	733
	db	11010000b
	dw	700
	db	10010000b
	dw	667
	db	11010000b
	dw	650
	db	10010000b
	dw	633
	db	01000000b
	dw	617
	db	10010000b
	dw	600
	db	10010000b
	dw	570
	db	10010000b
	dw	566
	db	01000000b
	dw	550
	db	10010000b
	dw	533
	db	11010000b
	dw	522
	db	10010000b
	dw	0
	db	10010000b

aA:
	db	'A', 0
aB:
	db	'B', 0
aE:
	db	'E', 0
aMHz:
	db	' MHz', 0

IntelMultTab:
	db	10, 6, 8, 4		; table for Intel P6 processor
	db	11, 7, 9, 5		; multipliers matching the bit
	db	1, 14, 16, 12		; fields in MSR 2Ah
	db	4, 15, 3, 13		;
	db	18, 22, 24, 1		; (all ratios are doubled)
	db	19, 23, 17, 1		;
	db	32, 1, 26, 20		; "1" represents a (probably)
	db	28, 1, 30, 21		; unused multiplier

KlamathMultTab:
	db	4, 6, 8, 4		; special table for Klamath cores
	db	1, 7, 9, 5
	db	1, 1, 1, 1
	db	4, 1, 3, 1

ViaMultTab1:
	db	10, 6, 8, 20		; table for Via processor multipliers
	db	11, 7, 9, 22		; valid for CPUIDs 066xh and 067xh
	db	18, 14, 16, 12
	db	24, 15, 26, 13

ViaMultTab2:
	db	10, 6, 8, 20		; table for Via processor multipliers
	db	11, 7, 9, 19		; valid for CPUID 068xh
	db	18, 14, 16, 12
	db	24, 15, 17, 13
	db	4, 22, 24, 4
	db	27, 23, 25, 21
	db	26, 30, 32, 28
	db	4, 31, 4, 29

ViaMultTab3:
	db	10, 32, 4, 20		; table for Via processor multipliers
	db	11, 4, 4, 19		; valid for CPUID 068xh
	db	18, 14, 16, 12
	db	24, 15, 17, 13
	db	18, 22, 24, 20
	db	27, 23, 25, 21
	db	26, 30, 32, 28
	db	24, 31, 4, 29

%ifdef HARDWARE_MONITOR
%ifdef PATCH_MINUS_5V_RAIL
aNA:
	db	'  N/A       '
%endif					; PATCH_MINUS_5V_RAIL
%endif					; HARDWARE_MONITOR

;----------------------------------------------------------------------------
; code
;----------------------------------------------------------------------------

; =============== S U B R O U T I N E =======================================


checkIntel:
	pushad
	xor	eax, eax
	cpuid
	cmp	ebx, 'Genu'
	jnz	short .end
	cmp	edx, 'ineI'
	jnz	short .end
	cmp	ecx, 'ntel'

.end:
	popad
	retn


; =============== S U B R O U T I N E =======================================


checkVia:
	pushad
	xor	eax, eax
	cpuid
	cmp	ebx, 'Cent'
	jnz	short .end
	cmp	edx, 'aurH'
	jnz	short .end
	cmp	ecx, 'auls'

.end:
	popad
	retn


; =============== S U B R O U T I N E =======================================


determineCPUType:
	push	ds
	push	si
	push	di
	push	cs
	pop	ds
	xor	eax, eax		; get CPU manufacturer
	cpuid
	mov	si, ManufacturerIDs

.checkManufacturer:
	cmp	word [si], 0FFFFh	; end of list reached?
	jz	short .useFallback
	cmp	ebx, [si+2]		; compare CPUID manufacturer ID to
	jnz	short .nextManufacturer	; list value
	cmp	ecx, [si+6]
	jnz	short .nextManufacturer
	cmp	edx, [si+0Ah]
	jz	short .checkModel	; manufacturer found, check CPU model

.nextManufacturer:
	add	si, 0Eh			; no match, go to next manufacturer
	jmp	short .checkManufacturer
;----------------------------------------------------------------------------

.checkModel:
	mov	si, [si]		; switch to model list
	mov	di, si
	mov	eax, 1			; get CPU model
	cpuid
	mov	ebx, eax		; split into family/model ...
	and	al, 0F0h
	and	bx, 0Fh			; ... and stepping

.nextModel:
	cmp	word [si], 0FFFFh	; end of list reached?
	jz	short .assumeDefault
	cmp	ax, [si]		; compare CPUID model to list value
	jz	short .modelFound	; model found
	add	si, 4
	jmp	short .nextModel	; no match, go to next model
;----------------------------------------------------------------------------

.assumeDefault:
	mov	si, IntelTypeMap

.modelFound:
	cmp	di, IntelTypeMap	; Intel CPU identified?
	jnz	short .checkVia
	cmp	ax, 650h		; CPUID 650h might be either a
	jnz	short .acceptType	; Deschutes or Covington core. A
	mov	ecx, 11Eh		; check for the presence of L2 cache
	rdmsr				; determines which core has been
	mov	dx, 1			; found. In case of a Covington core,
	shr	eax, 17h		; type 1 is selected.
	and	al, 1
	cmp	al, 1
	jz	short .storeType

.checkVia:
	cmp	di, ViaTypeMap		; Via CPU identified?
	jnz	short .acceptType
	cmp	ax, 670h		; CPUIDs 0677h and below belong to a
	jnz	short .acceptType	; Samuel 2 core, 0678h and above to
	cmp	bl, 8			; an Ezra core
	jb	short .acceptType
	mov	dx, 42h
	jmp	short .storeType
;----------------------------------------------------------------------------

.acceptType:
	mov	dx, [si+2]		; copy type from model list
	jmp	short .storeType
; ---------------------------------------------------------------------------

.useFallback:
	mov	dx, 1

.storeType:
	mov	ax, dx

.end:
	pop	di
	pop	si
	pop	ds
	retn


; =============== S U B R O U T I N E =======================================


getFSB:
	push	es
	mov	ax, 0
	mov	es, ax
	mov	al, [es:FSBIdx]
	pop	es
	and	al, 0F0h
	lea	si, [FSBTable]		; search matching FSB index

.nextFSBIdx:
	cmp	al, [cs:si+1]
	jz	short .foundFSB
	add	si, 2
	jmp	short .nextFSBIdx
; ---------------------------------------------------------------------------

.foundFSB:
	movzx	ax, [cs:si]		; get FSB value
	retn


; =============== S U B R O U T I N E =======================================


getMult:
	push	si
	push	edx
	push	ebx
	mov	bl, 2Fh			; See below: Via cores from (at
	call	checkVia		; least) Nehemiah onwards use the
	jz	short .getMultiplier	; additional bit. In previous cores
					; the bit is reserved, but should
					; return 0 according to datasheets,
					; so using it anyway shouldn't
					; hurt(?).
	mov	eax, 1
	cpuid
	and	ax, 0FFFh
	mov	bl, 0Fh			; Prepare the bit mask for the
	cmp	ax, 683h		; multiplier bit field in MSR 2Ah:
	jb	short .checkTualatin	; CPUs from 0683h onwards have an
	cmp	ax, 68Fh		; additional bit for the multiplier.
	ja	short .checkTualatin	; (This does not apply to 069xh,
	mov	bl, 2Fh			; which are mobile procesors anyway.)

.checkTualatin:
	and	ax, 0FF0h		; Tualatin cores also use the
	cmp	ax, 6B0h		; additional bit.
	jnz	short .getMultiplier
	mov	bl, 2Fh

.getMultiplier:
	mov	ecx, 2Ah		; get the multiplier bit field from
	rdmsr				; MSR 2Ah
	shr	eax, 16h
	and	al, bl
	test	al, 20h			; stuff the bits together (the
	jz	short .codeToMult	; additional bit is separated from
	and	al, 0DFh		; the rest)
	or	al, 10h			

.codeToMult:
	xor	ah, ah
	call	getMultTab		; get the multiplier table matching
	add	si, ax			; the processor type and find the
	mov	al, [cs:si]		; actual mutiplier
	pop	ebx
	pop	edx
	pop	si
	retn


%if FEATURE_LEVEL >= 3

; =============== S U B R O U T I N E =======================================


getMultFar:
	call	getMult
	retf

%endif					; FEATURE_LEVEL


; =============== S U B R O U T I N E =======================================


getMultTab:
	push	ax
	call	determineCPUType

	test	ax, 40h
	jnz	short .via

	mov	si, KlamathMultTab	; Klamath cores use multiplier 2.0
	cmp	ax, 02h			; where newer Intel processor use
	jz	short .end		; mulitplier 5.0
	mov	si, IntelMultTab
	jmp	short .end
; ---------------------------------------------------------------------------
	
.via:
	mov	si, ViaMultTab2		; newer Via processors use extended
	cmp	ax, 43h			; multiplier tables, otherwise we
	jz	short .end		; take the short one
	mov	si, ViaMultTab3
	cmp	ax, 44h
	jz	short .end
	mov	si, ViaMultTab1

.end:
	pop	ax
	retn


; =============== S U B R O U T I N E =======================================


calcFSB:
	push	bx
	call	getMult			; determine the CPU's multiplier
	mov	bl, al
	mov	ax, [bp+cpuFreq]	; divide frequency by multiplier to
	shl	ax, 1			; get a rough estimate of the FSB
	inc	ax
	div	bl
	mov	si, FSBTable

.nextFSB:
	cmp	byte [cs:si], 0		; find best FSB match
	jz	short .notFound
	mov	ah, [cs:si+1]
	cmp	al, [cs:si]
	jnb	short .acceptResult
	inc	si
	inc	si
	jmp	short .nextFSB
; ---------------------------------------------------------------------------

.notFound:
	xor	ah, ah

.acceptResult:
	pop	bx
	or	ah, ah
	jz	short .end
	mov	bl, ah

.end:
	retn


; =============== S U B R O U T I N E =======================================


printCPUSpeed0:
	call	getFSB			; fourth parameter: FSB speed
	push	ax

	mov	ax, word [bp+cpuFreq]	; third parameter: CPU frequency
	push	ax

	mov	eax, 1			; second parameter: brand index
	cpuid
	push	bx

	call	determineCPUType	; first parameter: CPU type
	push	ax

	call	printCPUSpeed
	add	esp, 8			; clean up parameters

	retn


; =============== S U B R O U T I N E =======================================


printCPUSpeed:
	.type	equ 2
	.brand	equ 4
	.freq	equ 6
	.fsb	equ 8

	mov	ax, [esp+.freq]		; get CPU speed

	push	ax
	push	0E000h
	push	.afterConvert
	push	retFarF000
	push	itoa4			; convert speed to string (length 4)
	jmp	0F000h:retNearF000
; ---------------------------------------------------------------------------

.afterConvert:
	pop	ax
	inc	byte [bp+screenCol]
	cmp	ax, 1000		; speed only 3 characters long?
	jnb	short .printSpeed
	dec	byte [bp+screenCol]

.printSpeed:
	push	0E000h
	push	.afterPrint
	push	retFarF000
	push	printNumber		; print speed
	jmp	0F000h:retNearF000
; ---------------------------------------------------------------------------

.afterPrint:
	cmp	word [esp+.type], 8	; check for various core types
	jnz	short .checkCoppermine	; and call special handling routines
	call	printTualatin		; for each one
	jmp	short .printUnit
; ---------------------------------------------------------------------------

.checkCoppermine:
	cmp	word [esp+.type], 6
	jnz	short .checkKatmai
	call	printCoppermine
	jmp	short .printUnit
; ---------------------------------------------------------------------------

.checkKatmai:
	cmp	word [esp+.type], 5
	jnz	short .checkMendocino
	call	printKatmai
	jmp	short .printUnit
; ---------------------------------------------------------------------------

.checkMendocino:
	cmp	word [esp+.type], 4
	jnz	short .checkVia
	call	printMendocino
	jmp	short .printUnit
; ---------------------------------------------------------------------------

.checkVia:
	test	word [esp+.type], 40h
	jz	short .printUnit
	call	printVia

.printUnit:
	mov	si, aMHz		; " MHz"
	call	printCtrlString		; finish by printing the "MHz" suffix
	retn


; =============== S U B R O U T I N E =======================================


printTualatin:
	.type	equ 2
	.brand	equ 4
	.freq	equ 6
	.fsb	equ 8

	call	getL2CacheSize		; Pentium III-S CPUs need no special
	cmp	al, 6			; discriminator
	jz	short .end
	cmp	word [esp+.freq], 1133	; All other Tualatin cores between
	ja	short .end		; 1000 and 1133 MHz get an "A"
	cmp	word [esp+.freq], 1000	; suffix. (The Tualatin-based Celeron
	jb	short .end		; at 900 MHz should have a suffix,
					; but doesn't have one.)
	mov	si, aA			; "A"
	call	printCtrlString

.end:
	retn


; =============== S U B R O U T I N E =======================================


printCoppermine:
	.type	equ 2
	.brand	equ 4
	.freq	equ 6
	.fsb	equ 8

	test	byte [esp+.brand], 1	; check for the Coppermine-based
	jz	short .checkPentium	; Celeron 533A
	cmp	word [esp+.freq], 533
	jnz	short .end
	mov	si, aA			; "A"
	call	printCtrlString
	jmp	short .end
; ---------------------------------------------------------------------------

.checkPentium:
	cmp	word [esp+.fsb], 100	; Coppermines at 100 MHz FSB between
	jnz	short .checkFSB133	; 500 and 600 MHz are distinguished
	cmp	word [esp+.freq], 600	; from their Katmai counterparts by
	ja	short .end		; an "E" suffix
	cmp	word [esp+.freq], 500
	jb	short .end
	mov	si, aE			; "E"
	call	printCtrlString
	jmp	short .end
; ---------------------------------------------------------------------------

.checkFSB133:
	cmp	word [esp+.fsb], 133	; Coppermines at 133 MHz FSB between
	jnz	short .end		; 533 and 600 MHz are suffixed with
	cmp	word [esp+.freq], 1000	; "EB". The 1000 MHz processor only
	jz	short .printB		; gets a "B" as it has no Katmai
	cmp	word [esp+.freq], 800	; counterpart. The latter is also
	jz	short .printE		; true for the 800 MHz processor, but
	cmp	word [esp+.freq], 533	; it gets the full suffix "EB"
	jb	short .end		; anyway.
	cmp	word [esp+.freq], 600
	ja	short .end

.printE:
	mov	si, aE			; "E"
	call	printCtrlString

.printB:
	mov	si, aB			; "B"
	call	printCtrlString

.end:
	retn


; =============== S U B R O U T I N E =======================================


printKatmai:
	.type	equ 2
	.brand	equ 4
	.freq	equ 6
	.fsb	equ 8

	cmp	word [esp+.fsb], 133	; Katmai-based processors at 133 MHz
	jnz	short .end		; FSB between 533 and 600 MHz are
	cmp	word [esp+.freq], 533	; suffixed with "B" to distinguish
	jb	short .end		; them from their 100 MHz FSB
	cmp	word [esp+.freq], 600	; counterparts.
	ja	short .end
	mov	si, aB			; "B"
	call	printCtrlString

.end:
	retn


; =============== S U B R O U T I N E =======================================


printMendocino:
	.type	equ 2
	.brand	equ 4
	.freq	equ 6
	.fsb	equ 8

	cmp	word [esp+.freq], 300	; check for the Mendocino-based
	jnz	short .end		; Celeron 300A
	mov	si, aA			; "A"
	call	printCtrlString

.end:
	retn


; =============== S U B R O U T I N E =======================================


printVia:
	.type	equ 2
	.brand	equ 4
	.freq	equ 6
	.fsb	equ 8

	cmp	word [esp+.type], 40h	; all C3 processors except Samuel
	jz	.end			; cores (CPUID 66xh) get an "A"
					; suffix
	mov	si, aA			; "A"
	call	printCtrlString

.end:
	retn


%ifdef HARDWARE_MONITOR

; =============== S U B R O U T I N E =======================================


getCPUVoltageRange:
	push	ax
	push	cx
	push	dx
	call	determineCPUType

	mov	bx, 8970h		; Covington, Deschutes, Mendocino,
	cmp	ax, 01h			; Katmai: ~ 2.0 V +/- 0.2 V
	jz	short .end
	cmp	ax, 03h
	jz	short .end
	cmp	ax, 04h
	jz	short .end
	cmp	ax, 05h
	jz	short .end

	mov	bx, 735Ah		; Coppermine: ~ 1.65 V +/- 0.2 V
	cmp	ax, 06h
	jz	short .end

	mov	bx, 6550h		; Tualatin: ~ 1.45 V +/- 0.17 V
	cmp	ax, 08h
	jz	short .end

	mov	bx, 0E6B6h		; Pentium Pro: ~ 3.3 V +/- 0.38 V
	cmp	ax, 09h			; (models vary from 3.1 V to 3.5 V)
	jz	short .end

	mov	bx, 0A88Fh		; OverDrive: ~ 2.5 V +/- 0.2 V
	cmp	ax, 0Ah			; (primary Vcc)
	jz	short .end

	mov	bx, 836Ah		; C5A/Samuel: ~ 1.9 V +/- 0.2 V
	cmp	ax, 40h			; (models vary from 1.8 V to 2.0 V)
	jz	short .end

	mov	bx, 6F56h		; C5B/Samuel 2: ~ 1.575 V +/- 0.2 V
	cmp	ax, 41h			; (models vary from 1.5 V to 1.65 V)
	jz	short .end

	mov	bx, 5F4Ah		; C5C/Ezra, C5M/C5N/Ezra-T,
	cmp	ax, 42h			; C5XL/Nehemiah:~ 1.35 V +/- 0.17 V
	jz	short .end		; (Nehemiah models vary from 1.25 V
	cmp	ax, 43h			; to 1.45 V)
	jz	short .end
	cmp	ax, 44h
	jz	short .end

	mov	bx, 0BBA2h		; Klamath/default: ~ 2.8 V +/- 0.2 V

.end:
	pop	dx
	pop	cx
	pop	ax
	retn


%ifdef HWMON_W83781D

; =============== S U B R O U T I N E =======================================


configHWMon:
	pusha
	call	configHWMon1
	mov	al, 4Bh
	call	readHWMon
	and	al, 11110011b
	or	al, 11000000b		; set fan 3 divisor to 1/8
	mov	ah, al
	mov	al, 4Bh
	call	writeHWMon
	popa
	retn

%endif					; HWMON_W83781D


%ifdef PATCH_MINUS_5V_RAIL

; =============== F U N C T I O N   C H U N K ===============================


checkMinus5VNA:
	cmp	ch, 1
	jz	afterVoltCheck

	cmp	ch, 6			; patch for -5 V: If the reading is
	jnz	resumeVoltCheck		; very low, we assume that no -5 V
	cmp	al, 5			; rail is present
	ja	resumeVoltCheck
	jmp	afterVoltCheck


; =============== S U B R O U T I N E =======================================


setItemStatusF:
	push	0E000h
	push	.end
	push	retFarF000
	push	setItemStatus
	jmp	0F000h:retNearF000
;----------------------------------------------------------------------------

.end:
	retn


; =============== F U N C T I O N   C H U N K ===============================


printMinus5V:
	push	ax
	mov	bx, minus5VMenu		; enable the -5 V menu item first
	xor	ax, ax
	call	setItemStatusF

	pop	cx			; get the reading for the -5 V rail
	mov	al, 20h
	add	al, ch
	call	readHWMon
	push	cx
	cmp	al, 5			; if the reading is very low, we
	pop	ax			; assume that no -5 V rail is
	ja	printVoltage		; present; otherwise, continue with
					; the normal display logic

	mov	si, aNA			; set "N/A" as display value
	push	di
	push	ds
	push	es
	mov	dx, 0E000h
	mov	ds, dx
	mov	dx, 0F000h
	mov	es, dx
	mov	cx, 3
	rep movsd
	pop	es
	pop	ds
	pop	di

	mov	bx, minus5VMenu		; select "N/A" entry for display
	mov	dx, 1
	push	0E000h
	push	.afterSelect
	push	retFarF000
	push	setItemSelect
	jmp	0F000h:retNearF000
;----------------------------------------------------------------------------

.afterSelect:
	mov	bx, minus5VMenu		; disable menu item
	mov	ax, 4
	call	setItemStatusF

	clc
	retf

%endif					; PATCH_MINUS_5V_RAIL
%endif					; HARDWARE_MONITOR


; =============== S U B R O U T I N E =======================================


getCPUBrand0:
	mov	eax, 1			; second parameter: brand index
	cpuid
	push	bx

	call	determineCPUType	; first parameter: CPU type
	push	ax

	call	getCPUBrand
	add	esp, 4			; clean up parameters

	retn


; =============== S U B R O U T I N E =======================================


getCPUBrand:
	.type	equ 2
	.brand	equ 4

	mov	ax, cs
	mov	ds, ax
					; Here, we deduce the brand string
					; of the processor according to the
					; coarse typing done in
					; determineCPUType. We try to get the
					; "easy" ones out of the way first
					; and do the more complicated ones
					; (the various flavors of
					; Pentium IIIs and Pentium III-based
					; Celerons) at the end.

	mov	si, aViaC3		; "VIA C3"
	test	word [esp+.type], 40h
	jnz	short .end

	mov	si, aCeleron		; "Intel(R) Celeron(TM)"
	cmp	word [esp+.type], 1	; "Type 1" are Intel processors
	jz	short .end		; without L2 cache (Covington core).
					; These are always Celerons.

	cmp	word [esp+.type], 4	; Mendocino cores are always Celerons
	jz	short .end

	mov	si, aPentiumPro		; "Intel(R) Pentium(R) Pro"
	cmp	word [esp+.type], 9
	jz	short .end

	mov	si, aPentium2		; "Intel(R) Pentium(R) II"
	cmp	word [esp+.type], 0Ah	; OverDrive processors get a "-ODP"
	jz	short .end		; suffix

	mov	si, aPentium3		; "Intel(R) Pentium(R) III"
	cmp	word [esp+.type], 5	; Katmai cores are always Pentium
	jz	short .end		; IIIs

	cmp	word [esp+.type], 8	; Tualatin cores might be Celerons,
	jz	short .checkP3S		; Pentium IIIs or Pentium III-Ss. We
					; check the latter first.

	cmp	word [esp+.type], 6	; Coppermine cores might be Celerons
	jz	short .checkP3Celeron	; or Pentium IIIs.

	jmp	short .default
; ---------------------------------------------------------------------------

.checkP3S:
	call	getL2CacheSize		; Tualatin core: Check the cache size
	cmp	al, 6			; (in 8 KB units: 2^6 * 8 KB = 512
	jnz	short .checkP3Celeron	; KB). A Tualatin core with 512 KB of
					; L2 cache is always a Pentium III-S,
					; otherwise it is a normal Pentium
					; III or a Celeron.
	mov	si, aPentium3S		; "Intel(R) Pentium(R) III-S"
	jmp	short .end
; ---------------------------------------------------------------------------

.checkP3Celeron:
	test	byte [esp+.brand], 1	; These families support the CPUID
					; brand index feature, so we try to
					; identify them as Celerons.
	jz	short .end
	mov	si, aCeleron		; "Intel(R) Celeron(TM)"
	jmp	short .end
; ---------------------------------------------------------------------------

.default:
	mov	si, aPentium2		; "Intel(R) Pentium(R) II"
					; If no other type could be
					; identified, we are looking at a
					; regular Klamath or Deschutes core,
					; i.e. a Pentium II.

.end:
	retn


; =============== F U N C T I O N   C H U N K ===============================


getECC:
	cmp	ax, 630h		; check for Pentium II-based cores
	jb	endGetECC		; (without OverDrive processors)
	cmp	ax, 660h
	jmp	resumeGetECC


%ifdef PATCH_SYMBIOS

; =============== S U B R O U T I N E =======================================


searchSymbios:
	pushad
	cmp	al, 2			; if setup option is "Enabled", force
	jz	.found			; load of option ROM, otherwise start
	mov	ecx, 80000000h		; searching for a matching adapter

.continue:
	xor	cl, cl
	call	queryPCI		; query PCI configuration space for
	cmp	ax, 1000h		; vendor 1000h: Symbios(/LSI/NCR)
	jnz	short .toNextBlock
	mov	cl, 2
	call	queryPCI		; query PCI configuration space for
	cmp	ax, 5			; device id 0001h, 0005h or 0006h
	jz	short .found
	cmp	ax, 6
	jz	short .found
	cmp	ax, 1
	jnz	short .toNextBlock

.found:
	mov	si, symbiosModuleID
	call	loadROMModule		; search and unpack ROM module
	jb	short .end
	call	unknown15
	jmp	short .end
; ---------------------------------------------------------------------------

.toNextBlock:
	add	ch, 8
	jnb	short .continue

.end:
	popad
	retn


; =============== S U B R O U T I N E =======================================


loadPCIExt:
%ifdef ONBOARD_NIC
	mov	si, onboardNICMenu
	call	getSetupOption
	or	al, al
	jz	short .afterNIC
	call	loadNICROM
%endif					; ONBOARD_NIC

.afterNIC:
%ifdef ONBOARD_ATA
	mov	si, onboardATAMenu
	call	getSetupOption
	or	al, al
	jz	short .afterATAFirst
	call	searchATA
%elifdef ONBOARD_SCSI
	mov	si, onbSCSIMenu
	call	getSetupOption
	test	al, 1
	jnz	short .afterSCSIFirst
%endif					; ONBOARD_ATA / ONBOARD_SCSI

.afterATAFirst:
%ifndef ONBOARD_SCSI
	mov	si, symbiosMenu		; get setup option for Symbios option
	call	getSetupOption		; ROM
	test	al, 1
	jnz	short .afterSCSIFirst	; if disabled, skip initialization
%else					; ONBOARD_SCSI
	push	ax			; save menu choice
	mov	si, onbSCSI1stMenu	; check whether onboard SCSI should
	call	getSetupOption		; have preference
	or	al, al
	pop	ax			; restore menu choice
	jz	short .lookForSymbios
	pushad				; if not, load the onboard ROM
	mov	ebx, 40040000h		; now
	call	uncompress
	popad
	call	searchOnbSCSI		; search for SCSI host adapter
%endif					; ONBOARD_SCSI

.lookForSymbios:
	call	searchSymbios		; search for Symbios adapter and
					; load ROM

.afterSCSIFirst:
	mov	ecx, 80000000h
	call	unknown12
	mov	ecx, 80000000h
	call	unknown13

%ifdef ONBOARD_ATA
	mov	si, onboardATAMenu
	call	getSetupOption
	or	al, al
	jnz	short .pciSnoop
	call	searchATA
%elifdef ONBOARD_SCSI
	mov	si, onbSCSIMenu
	call	getSetupOption
	test	al, 1
	jnz	short .pciSnoop
	mov	si, onbSCSI1stMenu
	call	getSetupOption
	or	al, al
	jnz	short .pciSnoop
	pushad
	mov	ebx, 40040000h		; load the onboard ROM
	call	uncompress
	popad
	call	searchOnbSCSI
%endif					; ONBOARD_ATA / ONBOARD_SCSI

.pciSnoop:
	mov	si, pciSnoopMenu
	call	getSetupOption
	or	al, al
	jz	short .end
	mov	ecx, 800000h
	mov	cx, [bp+0EAh]
	cmp	cx, 0
	jz	short .end
	shl	ecx, 8
	mov	cl, 4
	call	queryPCI
	test	al, 3
	jz	short .end
	or	al, 20h
	call	unknown14

.end:
	retn

%endif					; PATCH_SYMBIOS


; =============== F U N C T I O N   C H U N K ===============================


muCheckIntel:
	call	checkIntel
	jnz	skipMU
	mov	ebx, 40010000h
	jmp	resumeMU


; =============== F U N C T I O N   C H U N K ===============================


printVarNumber:
	mov	cx, printNumLen4
	cmp	ax, 1000
	jge	short .doPrint

	mov	cx, printNumLen3
	cmp	ax, 100
	jge	short .doPrint

	mov	cx, printNumLen2

.doPrint:
	push	0E000h
	push	.afterPrint
	push	retFarF000
	push	cx
	jmp	0F000h:retNearF000
; ---------------------------------------------------------------------------

.afterPrint:
	jmp	resumePrintL2


; =============== F U N C T I O N   C H U N K ===============================


configureCaches:
	mov	eax, 1			; only perform check of L2 cache
	cpuid				; address range for cores from
	and	ax, 0FFFh		; Klamath onwards
	cmp	ax, 630h
	jb	l2RangeDone
	mov	ecx, 11Eh
	jmp	resumeCfgCaches


; =============== S U B R O U T I N E =======================================


controlL2Cache:
	pushad
	push	ax
	mov	eax, 1
	cpuid
	mov	esi, eax		; save CPUID for Via check
	cmp	al, 30h
	pop	ax
	jnb	short .checkVia
	mov	bl, al			; perform L2 cache control for
	mov	ecx, 119h		; Pentium Pro cores
	rdmsr
	and	eax, 0FFF7FFFFh
	cmp	bl, 0FFh
	jz	short .configure
	or	eax, 80000h
	jmp	short .configure
; ---------------------------------------------------------------------------

.checkVia:
	call	checkVia		; only perform L2 cache control for
	jnz	short .end		; Via cores from CPUID 67xh onwards
	cmp	si, 670h
	jb	short .end
	mov	bl, al
	mov	ecx, 1107h
	rdmsr
	and	eax, 0FFFFFEFFh
	cmp	bl, 0FFh
	jz	short .configure
	or	eax, 100h

.configure:
	wrmsr

.end:
	popad
	retn


%if FEATURE_LEVEL = 0

; =============== S U B R O U T I N E =======================================


getCacheDesc:
	push	0E000h
	push	.end
	push	retFarF000
	push	getCacheDscF000
	jmp	0F000h:retNearF000
;----------------------------------------------------------------------------

.end:
	retn

%endif					; FEATURE_LEVEL


; =============== S U B R O U T I N E =======================================


getL2CacheSize:
	push	ebx
	push	ecx
	push	edx
	push	esi
	call	checkVia		; Via CPU?
	jz	short .via
	mov	eax, 2			; no, get cache size "Intel style"
	cpuid
	call	getCacheDesc
	jnb	short .end
	mov	eax, ebx
	call	getCacheDesc
	jnb	short .end
	mov	eax, ecx
	call	getCacheDesc
	jnb	short .end
	mov	eax, edx
	call	getCacheDesc
	jnb	short .end
	jmp	short .none
; ---------------------------------------------------------------------------

.via:
	mov	eax, 1
	cpuid
	mov	esi, eax		; save CPUID
	cmp	si, 670h		; only cores from CPUID 67xh onwards
	jb	short .none		; have L2 cache
	mov	ecx, 1107h		; check whether L2 cache is disabled
	rdmsr
	test	ax, 100h
	jnz	short .none
	mov	eax, 80000000h		; Via CPUs store L2 cache size
	cpuid				; information in extended CPUID
	cmp	eax, 80000006h		; registers
	jb	short .none
	mov	eax, 80000006h
	cpuid
	shr	ecx, 19			; L2 cache size in bits 31:16 for
	cmp	si, 690h		; Nehemiah, but in 31:24 for
	jae	short .convert		; Samuel 2; others are unclear
	shr	ecx, 8

.convert:				; convert to logarithm as expected by
	xor	al, al			; caller

.loop:
	shr	cx, 1
	jz	short .end
	inc	al
	jmp	short .loop
; ---------------------------------------------------------------------------

.none:
	xor	al, al

.end:
	pop	esi
	pop	edx
	pop	ecx
	pop	ebx
	retn


; =============== S U B R O U T I N E =======================================


getL2CacheSzFar:
	call	getL2CacheSize
	retf


%ifdef ESCD_PORT_WRITE

; =============== F U N C T I O N   C H U N K ===============================


writeESCD:
	mov	eax, 1
	cpuid
	cmp	ax, 630h
	jb	afterB2Port
	mov	al, 0A4h
	mov	dx, 0B2h
	jmp	resumeWriteESCD

%endif					; ESCD_PORT_WRITE


; =============== F U N C T I O N   C H U N K ===============================


initPentiumPro:
	cmp	ax, 611h		; disable fast string operations for
	jnz	short .disStrBuf	; Pentium Pro stepping B0 (CPUID
	mov	ecx, 1E0h		; 611h)
	rdmsr
	and	eax, 2
	or	al, 4
	wrmsr
	jmp	initPProDone
; ---------------------------------------------------------------------------

.disStrBuf:
	cmp	ax, 619h		; disable streaming buffers for
	jmp	continuePPro		; Pentium Pro stepping sB1 (CPUID
					; 619h) -- errata 58 and 59
